import unittest

from test.boggle_test import BoggleTest
from test.scrabble_test import ScrabbleTest
from test.hangman_test import HangmanTest
from test.wordsearch_test import WordSearchTest

if __name__ == "__main__":
    unittest.main()



"""
    This module contains the HangmanSolver, used to list the remaining letters
    in a game of Hangman, in order of chance of being correct.
"""

from collections import defaultdict
import re
from string import ascii_uppercase
from itertools import product
from wordsolver import wordtools


class HangmanSolver(wordtools.WordHash):
    """
        Class for solving a game of Hangman.

        Attempt to guess the solution.

        Attributes:
        > _candidates (set) - list of words that satify the criteria
        > _length (int) - length of the solution
        > _correct (dict) - mapping of correct letters
        > _incorrect (iterable) - letters which are not in the word
        > _words_of_length (dict) - keys are word lengths, values are all words
            added with that length
    """

    def __init__(self, filename):
        super().__init__()
        self.candidates = set()
        self.length = 0
        self.correct = defaultdict(set)
        self.incorrect = defaultdict(set)
        self.attempt = ""
        self.words_of_length = defaultdict(set)
        self._upload_words(filename)

    def _upload_words(self, collection):
        """Upload the Dictionary words from a text file 'filename'. When
        solving, the words from this text file will be used for spell
        checking."""
        self.clear_words()
        for word in self._validate_collection(collection):
            if not word.isalpha():
                continue
            self.add_word(word)
            self.words_of_length[len(word)].add(word)

    def solve(self, attempt, incorrect):
        """
            Returns a set of all words which fit the criteria.

            Parameters:
            > attempt (iterable) - the current guess of the solution. All
                unknown letters must be filled in with an non-alphabetical
                character.
            > incorrect (iterable) - letters which are not in the word

            Returns:
            > (set) - all words which fit the criteria
        """
        attempt, incorrect = self._validate_parameters(attempt, incorrect)
        self._reduce_candidates(attempt, incorrect)
        return self.candidates

    def guess_distribution(self, attempt, wrong):
        """
            Returns the probability for each letter to be in the word.

            Parameters:
            > attempt (iterable) - the current guess of the solution. All
                unknown letters must be filled in with an non-alphabetical
                character.
            > wrong (iterable) - letters which are not in the word

            Returns:
            > (list) - each element is in the form (letter, probability)
        """
        attempt, wrong = self._validate_parameters(attempt, wrong)
        self._reduce_candidates(attempt, wrong)

        tally = dict.fromkeys(set(ascii_uppercase) - self.incorrect.keys(), 0)

        for candidate in self.candidates:
            for letter in set(candidate) & tally.keys():
                tally[letter] += 1/len(self.candidates)

        return sorted(tally.items(), key=lambda x: -x[1])

    def _reduce_candidates(self, attempt, wrong):
        """
            Filters out all candidates in self.candidates which do not reach
            the specification.

            Parameters:
            > attempt (iterable) - current guess
            > wrong (iterable) - wrong letters
        """
        self.length = len(attempt)
        self.candidates = self.words_of_length[self.length]

        self.correct.clear()
        self.incorrect.clear()

        for pos, let in enumerate(attempt):
            if not let in ascii_uppercase:
                continue
            self.correct[let].add(pos)
            self.incorrect[let] |= (set(range(self.length)) - {pos})

        for pos, let in product(range(self.length), wrong):
            if not let in ascii_uppercase:
                continue
            self.incorrect[let].add(pos)

        self._filter_correct_letters(self.correct)
        self._filter_incorrect_letters(self.incorrect)

    def _filter_correct_letters(self, requirements):
        """
            Filters out all candidates in self.candidates which do not reach
            the specification.

            The specification is that for each (letter, positions) pair in
            'requirements', a candidate must have the letter at all the indicies
            in positions.

            Parameters:
            > requirements (dict) - mapping letters (str) and positions (list)
        """
        self.candidates = self.candidates.intersection(
            *[
                self.lookup(self.length, let, pos)
                for let, pos in self._yield_requirements(requirements)
            ]
        )

    def _filter_incorrect_letters(self, requirements):
        """
            Filters out all candidates in self.candidates which do not reach
            the specification.

            The specification is that for each (letter, positions) pair in
            'requirements', a candidate must not have the letter in any of the
            indicies in positions.

            Parameters:
            > requirements (dict) - mapping letters (str) and positions (list)
        """
        self.candidates = self.candidates.difference(
            *[
                self.lookup(self.length, let, pos)
                for let, pos in self._yield_requirements(requirements)
            ]
        )

    @staticmethod
    def _yield_requirements(requirements):
        """Yields the letter and positions in requirements."""
        for letter, positions in requirements.items():
            for position in positions:
                yield (letter, position)

    @staticmethod
    def _validate_collection(collection):
        """Validate the collection parameter. Returns the words extracted."""

        # Return collection if it is already a list/set
        if isinstance(collection, (list, set)):
            return [word.upper() for word in collection]

        # Opens file and return lists of words in the file
        if isinstance(collection, str):
            with open(collection, "r") as file_text:
                text = file_text.read()
            return re.findall(r"[\w']+", text.upper())

        # Else an incorrect type has been received
        raise TypeError(
            "invalid value for 'collection' parameter. "
            "Expected list, set or str, received '%s'." % type(collection).__name__
        )

    @staticmethod
    def _validate_parameters(attempt, wrong):

        # Check that attempt is a list or str
        if isinstance(attempt, str):
            attempt = list(attempt)
        elif not isinstance(attempt, list):
            raise TypeError(
                "invalid value for 'attempt' parameter. "
                "Expected list, set or str received '%s'." % type(attempt).__name__
            )

        # Check that attempt elements are strings
        element_lengths = set()
        for index, element in enumerate(attempt):
            if not isinstance(element, str):
                raise TypeError(
                    "invalid value for 'attempt' parameter at index %i. "
                    "Expected str, received '%s'." % (index, type(element).__name__)
                )
            attempt[index] = element.upper()
            element_lengths.add(len(element))

        # Check each element of attempt is a single character
        if element_lengths != {1}:
            raise ValueError(
                "invalid value for 'attempt' parameter. "
                "Not all strings are characters."
            )

        # Check that wrong is a list set or str
        if isinstance(wrong, (list, str)):
            wrong = set(wrong)
        elif not isinstance(wrong, set):
            raise TypeError(
                "invalid value for 'wrong' parameter. "
                "Expected list, set or str received '%s'." % type(attempt).__name__
            )

        # Check that wrong is a set of strings
        element_lengths = set()
        for index, element in enumerate(wrong):
            if not isinstance(element, str) or not element.isalpha():
                raise TypeError(
                    "invalid value for 'wrong' parameter at index %i. "
                    "Expected alphabet character, received '%s'." % (index, type(element).__name__)
                )
            element_lengths.add(len(element))
        wrong = {letter.upper() for letter in wrong}

        # Check each element of wrong is a single character (if non-empty)
        if wrong and element_lengths != {1}:
            raise ValueError(
                "invalid value for 'wrong' parameter. "
                "Not all strings are characters."
            )

        return attempt, wrong



"""
    The module contains the WordSearchSolver, used to solve a word search.
"""

from enum import Enum
from itertools import chain
import re

from wordsolver import wordtools


class Direction(Enum):
    """Compass directions."""
    NORTH = 1
    NORTH_EAST = 2
    EAST = 3
    SOUTH_EAST = 4
    SOUTH = 5
    SOUTH_WEST = 6
    WEST = 7
    NORTH_WEST = 8


class WordSearchSolver(wordtools.WordTree):
    """
        Class for solving a Word Search.

        Attributes:
        > grid (list) - 2d matrix representing letters and positions in a word
            search grid
        > width (int) - width of the word search grid
        > height (int) - height of the word search grid
    """

    DIRECTIONS = {
        "N": Direction.NORTH,
        "NE": Direction.NORTH_EAST,
        "E": Direction.EAST,
        "SE": Direction.SOUTH_EAST,
        "S": Direction.SOUTH,
        "SW": Direction.SOUTH_WEST,
        "W": Direction.WEST,
        "NW": Direction.NORTH_WEST
    }

    def __init__(self, collection):
        super().__init__()
        self.grid = None
        self.width = 0
        self.height = 0
        self._yield_slice_funcs = dict()
        self._setup(collection)

    def _setup(self, collection):
        """
            Upload the hidden words from a text file. The hidden words will be
            seaarched for when solving. Words can be uploaded from a file and/or
            from a list.

            Parameters:
            > filename (str) - name of the text file
        """
        self.clear_words()
        for word in self._validate_collection(collection):
            self.add_word(word.upper())

        self._yield_slice_funcs = {
            Direction.NORTH: self._yield_slices_north,
            Direction.NORTH_EAST: self._yield_slices_north_east,
            Direction.EAST: self._yield_slices_east,
            Direction.SOUTH_EAST: self._yield_slices_south_east,
            Direction.SOUTH: self._yield_slices_south,
            Direction.SOUTH_WEST: self._yield_slices_south_west,
            Direction.WEST: self._yield_slices_west,
            Direction.NORTH_WEST: self._yield_slices_north_west
        }

    def solve(self, grid, directions=None):
        """
            Find the positions of all hidden words.

            Parameters:
            > grid (list) - 2d matrix of letters, representing the word search
                grid to be solved
            > directions (set) - compass directions to search for hidden words

            Returns:
            > solutions (list) - a list of tuples of the following form:
                (word, start_pos, end_pos)
                - word (str) : hidden word
                - start_pos (int, int) : x, y position of the start of the word
                - end_pos (int, int) : x, y position of the end of the word
        """
        directions = directions or {"ALL"}

        grid = self._validate_grid(grid)
        directions = self._validate_directions(directions)

        self.width = len(grid[0])
        self.height = len(grid)
        self.grid = [
            [((col, row), grid[row][col].upper()) for col in range(self.width)]
            for row in range(self.height)
        ]

        solutions = list()
        for grid_slice in self._generate_slices(directions):
            for solution in self._get_solutions(grid_slice):
                solutions.append(solution)

        return solutions

    def _generate_slices(self, directions):
        """
            Generator function used to extract word search grid slices.

            Parameters:
            > direction (Direction) - the direction to take the slices in

            Yields:
            > grid_slice (list) - a grid slice of the word search grid, in the
                given 'direction'
        """
        generators = [
            func() for direction, func in self._yield_slice_funcs.items()
            if direction in directions
        ]

        for grid_slice in chain(*generators):
            yield grid_slice

    def _yield_slices_north(self):
        """Yield grid slices in a north direction."""
        for col in range(self.width):
            grid_slice = list()
            for row in range(self.height - 1, -1, -1):
                grid_slice.append(self.grid[row][col])
            yield grid_slice

    def _yield_slices_north_east(self):
        """Yield grid slices in a north-east direction."""
        for diag in range(self.width + self.height - 1):
            grid_slice = list()
            for col in range(max(0, 1-self.height+diag), min(self.width, diag+1)):
                grid_slice.append(self.grid[diag-col][col])
            yield grid_slice

    def _yield_slices_east(self):
        """Yield grid slices in an east direction."""
        for row in range(self.height):
            grid_slice = list()
            for col in range(self.width):
                grid_slice.append(self.grid[row][col])
            yield grid_slice

    def _yield_slices_south_east(self):
        """Yield grid slices in a south-east direction."""
        for diag in range(self.width + self.height - 1):
            grid_slice = list()
            for col in range(max(0, 1-self.height+diag), min(self.width, diag+1)):
                grid_slice.append(self.grid[col-diag + self.height - 1][col])
            yield grid_slice

    def _yield_slices_south(self):
        """Yield grid slices in a south direction."""
        for col in range(self.width):
            grid_slice = list()
            for row in range(self.height):
                grid_slice.append(self.grid[row][col])
            yield grid_slice

    def _yield_slices_south_west(self):
        """Yield grid slices in a south-west direction."""
        for diag in range(self.width + self.height - 1):
            grid_slice = list()
            for col in range(min(self.width, diag+1) - 1, max(0, 1-self.height+diag) - 1, -1):
                grid_slice.append(self.grid[diag-col][col])
            yield grid_slice

    def _yield_slices_west(self):
        """Yield grid slices in a west direction."""
        for row in range(self.height):
            grid_slice = list()
            for col in range(self.width-1, -1, -1):
                grid_slice.append(self.grid[row][col])
            yield grid_slice

    def _yield_slices_north_west(self):
        """Yield grid slices in a north-west direction."""
        for diag in range(self.width + self.height - 1):
            grid_slice = list()
            for col in range(min(self.width, diag+1) - 1, max(0, 1-self.height+diag) - 1, -1):
                grid_slice.append(self.grid[col-diag + self.height - 1][col])
            yield grid_slice

    def _get_solutions(self, grid_slice):
        """
            Generator function for finding words in a word search grid slice.

            Parameters:
            > grid_slice (list) - a slice of the word search grid

            Yields:
            > word (str) - a word found in the grid slice
        """
        for pos, begin_entry in enumerate(grid_slice):
            if not self.root.get_child(begin_entry[1]):
                continue
            node = self.root.get_child(begin_entry[1])
            for end_entry in grid_slice[pos+1:]:
                node = node.get_child(end_entry[1])
                if not node:
                    break
                if node.my_word:
                    yield (node.my_word, begin_entry[0], end_entry[0])

    @staticmethod
    def _validate_collection(collection):
        """Validate the collection parameter. Returns the words extracted."""

        # Return collection if it is already a list/set
        if isinstance(collection, (list, set)):
            return collection

        # Opens file and return lists of words in the file
        if isinstance(collection, str):
            with open(collection, "r") as file_text:
                text = file_text.read()
            return re.findall(r"[\w']+", text)

        # Else an incorrect type has been received
        raise TypeError(
            "invalid value for 'collection' parameter. "
            "Expected list, set or str, received '%s'." % type(collection).__name__
        )

    @staticmethod
    def _validate_grid(grid):
        """Validate the grid parameter. Returns the grid."""

        # Check grid is a list
        if not isinstance(grid, list):
            raise TypeError(
                "invalid value for 'grid' parameter. "
                "Expected list, received '%s'." % type(grid).__name__
            )

        # Check grid is a list of lists
        element_lengths = set()
        for index, element in enumerate(grid):
            if not isinstance(element, list):
                raise TypeError(
                    "invalid value for 'grid' parameter at index %i. "
                    "Expected list, received '%s'." % (index, type(element).__name__)
                )
            element_lengths.add(len(element))

        # Check all lists are the same size
        if len(element_lengths) > 1:
            raise ValueError(
                "invalid value for 'grid' parameter. "
                "Not all rows are the same size."
            )

        # Check each element in the lists are strings
        for index, element in enumerate(grid):
            for tile in element:
                if not isinstance(tile, str):
                    raise TypeError(
                        "invalid value for 'grid' parameter at index %i. "
                        "Not all elements are str, found '%s'" % (index, type(tile).__name__)
                    )
            grid[index] = [tile.upper() for tile in element]

        # Check that each element of grid is a single letter
        for index, element in enumerate(grid):
            for letter in element:
                if len(letter) != 1:
                    raise ValueError(
                        "invalid value for 'grid' parameter at index %i. "
                        "Cannot convert '%s' into a letter." % (index, letter)
                    )
        return grid

    def _validate_directions(self, directions):
        """Validate the directions parameter. Returns the directions."""

        # Check directions is a list or set
        if not isinstance(directions, (list, set)):
            raise TypeError(
                "invalid value for 'directions' parameter. "
                "Expected list or set , received '%s'." % type(directions).__name__
            )

        # Check each element in the directions are strings
        for index, element in enumerate(directions):
            if not isinstance(element, str):
                raise TypeError(
                    "invalid value for 'directions' parameter at index %i. "
                    "Not all elements are str, found '%s'" % (index, type(element).__name__)
                )

        # Check that directions are all valid compass directions or all
        directions = {dir.upper() for dir in directions}

        if not directions.issubset(self.DIRECTIONS.keys() | {"ALL"}):
            raise ValueError(
                "invalid value for 'directions' parameter. "
                "Permitted directions are N, NW, W, SW, S, SE, E, NE or ALL."
            )

        if "ALL" in directions:
            return self.DIRECTIONS.values()

        return {self.DIRECTIONS[direction] for direction in directions}



"""
    This module provides classes making the process of checking word
    spelling more efficient.
"""

from collections import defaultdict


class WordNode():
    """
        Node object for the WordTree data structure.

        Attributes:
        > letter (str) - value of the node
        > children (list) - list of child node objects
        > my_word (str) - if this node completes a word, my_word is
            assigned to that word, else it equals ""
    """

    def __init__(self, letter=""):
        self.letter = letter
        self.children = []
        self.my_word = ""

    def add_child(self, letter):
        """Creates a new child WordNode object, assigning it with the
        given letter."""
        child = WordNode(letter)
        self.children.append(child)
        return child

    def get_child(self, letter):
        """Returns the child WordNode with the given letter, if it
        exists."""
        for child in self.children:
            if child.letter == letter:
                return child
        return None

    def delete(self):
        """Deletes this WordNode object and all descendant child
        nodes."""
        for child in self.children:
            child.delete()
        del self


class WordTree():
    """
        Tree data structure object, used as an efficient method for
        checking word spelling. All words added form a path from the
        root node, with each node representing a letter of that word.

        Attributes:
        > root (WordNode) - root of the tree
        > word_set (set) - set of words added
    """

    def __init__(self):
        self.root = WordNode()
        self.word_set = set()

    def add_word(self, word):
        """Adds a word into the tree data structure."""
        current = self.root
        for letter in word:
            child = current.get_child(letter)
            if not child:
                child = current.add_child(letter)
            current = child
        current.my_word = word
        self.word_set.add(word)

    def clear_words(self):
        """Clears the all words from the tree data structure."""
        for child in self.root.children:
            child.delete()
        self.word_set = set()


class WordHash():
    """
        Hash table object, used as an efficient method for looking up
        words with particular letters in given positions. Reverse hash
        lookups can be used to return the set of all words with a given
        length and a particular letter at a given index. Words may
        belong to more that one set.

        Attributes:
        > hash_table (dict) - keys are hashes, values are sets of the
            words with that hash
    """

    def __init__(self):
        self.hash_table = defaultdict(set)

    def add_word(self, word):
        """Add the word to hash_table."""
        length = len(word)
        for position, letter in enumerate(word):
            hash_value = self._hasher(length, letter, position)
            self.hash_table[hash_value].add(word)

    def lookup(self, length, letter, position):
        """Returns the set of added words with the given length and with
        the given letter in the given position."""
        hash_value = self._hasher(length, letter, position)
        return self.hash_table[hash_value]

    def clear_words(self):
        """Clears the all words from the hash table."""
        self.hash_table = defaultdict(set)

    @staticmethod
    def _hasher(length, letter, position):
        """Returns a hash combining the given arguments."""
        return str(length) + letter + str(position)



"""Wordsolver Module"""

__version__ = "1.0.0"

from wordsolver.boggle import BoggleSolver
from wordsolver.scrabble import ScrabbleSolver, VALUE_STANDARD, PREMIUM_STANDARD, EMPTY_STANDARD
from wordsolver.hangman import HangmanSolver
from wordsolver.wordsearch import WordSearchSolver



"""
    This modules contains the BoggleSolver, used to find all the words in a
    Boggle board.
"""

import re

from collections import defaultdict
from itertools import product
from pythonds.basic.stack import Stack

from wordsolver import wordtools


class BoggleSolver(wordtools.WordTree):
    """
        Class for solving a Boggle board.

        Attributes:
        > word_paths (dict) - keys are the Dictionary words used for
            solving. The value for each word is a list of all possible
            board paths to attain that word. A board path is a list
            of board co-ordinates (x, y). The values of words that
            are not attainable on the board are empty lists.
        > substitutions (dict) - mapping for swapping sub-strings in
            words for a single character substitution.
            e.g. substitute = {"QU" : "Q"}. Needs to be all in caps.
    """

    SUBSTITUTIONS = {"QU": "Q"}

    def __init__(self, collection, min_length=3):
        super().__init__()
        self._setup(collection, min_length)
        self.word_paths = defaultdict(list)

    def _setup(self, collection, min_length):
        """
            Upload the Dictionary words from 'collection'. When solving,
            the words file will be used for finding words.

            Parameters:
            > collection (list/set/str) - can either be a list/set of strings
                or the filename of list of words
            > min_length (int) - the minimum length of words uploaded
        """
        self._validate_min_length(min_length)
        self.clear_words()
        for word in self._validate_collection(collection):
            if (not word.isalpha() or len(word) < min_length):
                continue
            substituted = self._apply_substitute(word.upper(), reverse=False)
            self.add_word(substituted)

    def solve(self, board, with_positions=False):
        """
            Iterates through all starting positions on the board, adding
            all possible words starting at that position.

            Parameters:
            > board (list) - 2d matrix representing a Boggle board, all
                entries in caps
            > with_positions (bool) - whether positions of the letters forming
                the words should be added to the solution

            Returns:
            > (list) - all words on the board, beginning with the longest
                word(s)
        """
        board = self._validate_board(board, with_positions)
        self.word_paths = defaultdict(list)
        for row, col in product(range(len(board)), range(len(board[0]))):
            self._iteration(board, row, col)
        if with_positions:
            return sorted(self.word_paths.items(), key=lambda x: -len(x[0]))
        return sorted(self.word_paths.keys(), key=len, reverse=True)

    def _apply_substitute(self, word, reverse=True):
        """If reverse is True, replace any single character
        substitutions in word with the sub-string substitution, found in
        self.SUBSTITUTIONS. If reverse is False, replace any sub-strings
        in word with the single character substitution."""
        for item in self.SUBSTITUTIONS.items():
            word = word.replace(item[reverse], item[not reverse])
        return word

    def _iteration(self, board, row, col):
        """
            Main iterative process. This method adds all possible board
            paths to self.word_paths, starting at position (row, col).
            The iterative process works as follows:
            1) starts with the path [(row, col)]
            2) checks all possible additions from the end of the path
            3) if an addition forms a word, the path is appended to the
            corresponding list in self.word_paths
            4) repeat from 2) until all possiblities from that branch
            have been exhausted
        """
        first_node = self.root.get_child(board[row][col])
        if not first_node:
            return
        stack = Stack()
        stack.push((first_node, [(row, col)]))
        while not stack.isEmpty():
            current = stack.pop()
            for (adj_row, adj_col) in self._get_adjacent(board, current[1]):
                adj_node = current[0].get_child(board[adj_row][adj_col])
                adj_path = current[1] + [(adj_row, adj_col)]
                if not adj_node:
                    continue
                if adj_node.my_word:
                    true_word = self._apply_substitute(adj_node.my_word)
                    self.word_paths[true_word].append(adj_path)
                stack.push((adj_node, adj_path))

    @staticmethod
    def _get_adjacent(board, path):
        """
            This method finds all possible candidate co-ordinates for
            additions to a path.

            Parameters:
            > board (list) - 2d matrix representing a Boogle Board
            > path (list) - board path

            Returns:
            > (list) - all board co-ordinates (x, y) which are adjacent
                to the last co-ordinate in path and are also not in path
        """
        adjacent = []
        row_range = max(0, path[-1][0]-1), min(len(board), path[-1][0]+2)
        col_range = max(0, path[-1][1]-1), min(len(board[0]), path[-1][1]+2)
        for adj_row, adj_col in product(range(*row_range), range(*col_range)):
            if (adj_row, adj_col) not in path:
                adjacent.append((adj_row, adj_col))
        return adjacent

    @staticmethod
    def _validate_collection(collection):
        """Validate the collection parameter. Returns the words extracted."""

        # Return collection if it is already a list/set
        if isinstance(collection, (list, set)):
            return collection

        # Opens file and return lists of words in the file
        if isinstance(collection, str):
            with open(collection, "r") as file_text:
                text = file_text.read()
            return re.findall(r"[\w']+", text)

        # Else an incorrect type has been received
        raise TypeError(
            "invalid value for 'collection' parameter. "
            "Expected list, set or str, received '%s'." % type(collection).__name__
        )

    @staticmethod
    def _validate_min_length(min_length):
        """Validate the min_length parameter."""

        # Check that value is an int
        if not isinstance(min_length, int):
            raise TypeError(
                "invalid value for 'min_length' parameter. "
                "Expected int, received '%s'." % type(min_length).__name__
            )

        # Check that value is in the correct range
        if min_length < 0:
            raise ValueError(
                "invalid value for 'min_length' parameter. "
                "Value needs to be non-negative"
            )

    def _validate_board(self, board, with_positions):
        """Validate the board parameter. Returns the board."""

        # Check board is a list
        if not isinstance(board, list):
            raise TypeError(
                "invalid value for 'board' parameter. "
                "Expected list, received '%s'." % type(board).__name__
            )

        # Check board is a list of lists
        element_lengths = set()
        for index, element in enumerate(board):
            if not isinstance(element, list):
                raise TypeError(
                    "invalid value for 'board' parameter at index %i. "
                    "Expected list, received '%s'." % (index, type(element).__name__)
                )
            element_lengths.add(len(element))

        # Check all lists are the same size
        if len(element_lengths) > 1:
            raise ValueError(
                "invalid value for 'board' parameter. "
                "Not all rows are the same size."
            )

        # Check each element in the lists are strings
        for index, element in enumerate(board):
            for tile in element:
                if not isinstance(tile, str):
                    raise TypeError(
                        "invalid value for 'board' parameter at index %i. "
                        "Not all elements are str, found '%s'" % (index, type(tile).__name__)
                    )
            board[index] = [
                self._apply_substitute(tile.upper(), reverse=False)
                for tile in element
            ]

        # Check that each element of board is a single letter
        for index, element in enumerate(board):
            for letter in element:
                if len(letter) != 1:
                    raise ValueError(
                        "invalid value for 'board' parameter at index %i. "
                        "Cannot convert '%s' into a letter." % (index, letter)
                    )

        # Check the with_positions paramter
        if not isinstance(with_positions, bool):
            raise TypeError(
                "invalid value for 'with_positions' parameter. "
                "Expected bool, received '%s'." % type(board).__name__
            )

        return board



"""
    This module contains the ScrabbleSolver, used to find all the possible
    words that can be played on a Scrabble board.
"""

import re
from string import ascii_lowercase, ascii_uppercase
from collections import defaultdict
import numpy as np

from wordsolver import wordtools


VALUE_STANDARD = {
    "A" : 1, "B" : 3, "C" : 3, "D" : 2, "E" : 1, "F" : 4, "G" : 2, "H" : 4,
    "I" : 1, "J" : 8, "K" : 5, "L" : 1, "M" : 3, "N" : 1, "O" : 1, "P" : 3,
    "Q" : 10, "R" : 1, "S" : 1, "T" : 1, "U" : 1, "V" : 4, "W" : 4, "X" : 8,
    "Y" : 4, "Z" : 10
}

PREMIUM_STANDARD = [
    ["T", "*", "*", "d", "*", "*", "*", "T", "*", "*", "*", "d", "*", "*", "T"],
    ["*", "D", "*", "*", "*", "t", "*", "*", "*", "t", "*", "*", "*", "D", "*"],
    ["*", "*", "D", "*", "*", "*", "d", "*", "d", "*", "*", "*", "D", "*", "*"],
    ["d", "*", "*", "D", "*", "*", "*", "d", "*", "*", "*", "D", "*", "*", "d"],
    ["*", "*", "*", "*", "D", "*", "*", "*", "*", "*", "D", "*", "*", "*", "*"],
    ["*", "t", "*", "*", "*", "t", "*", "*", "*", "t", "*", "*", "*", "t", "*"],
    ["*", "*", "d", "*", "*", "*", "d", "*", "d", "*", "*", "*", "d", "*", "*"],
    ["T", "*", "*", "d", "*", "*", "*", "D", "*", "*", "*", "d", "*", "*", "T"],
    ["*", "*", "d", "*", "*", "*", "d", "*", "d", "*", "*", "*", "d", "*", "*"],
    ["*", "t", "*", "*", "*", "t", "*", "*", "*", "t", "*", "*", "*", "t", "*"],
    ["*", "*", "*", "*", "D", "*", "*", "*", "*", "*", "D", "*", "*", "*", "*"],
    ["d", "*", "*", "D", "*", "*", "*", "d", "*", "*", "*", "D", "*", "*", "d"],
    ["*", "*", "D", "*", "*", "*", "d", "*", "d", "*", "*", "*", "D", "*", "*"],
    ["*", "D", "*", "*", "*", "t", "*", "*", "*", "t", "*", "*", "*", "D", "*"],
    ["T", "*", "*", "d", "*", "*", "*", "T", "*", "*", "*", "d", "*", "*", "T"]
]

EMPTY_STANDARD = [
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"],
    ["*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*", "*"]
]


class ScrabbleSolver(wordtools.WordHash):
    """
        Class for solving a Scrabble board.

        Attributes:
        > board (np.array) - 2d matrix representing a Scrabble board,
            entries in caps for normal tiles, lower case for blanks,
            empty spaces can be "", "*" or None
        > minor (np.array) - 2d matrix of scores obtained from minor
            words
        > rack (str) - letters in the rack, '#' for blanks
        > values (dict) - keys are possible tiles, values are the values
        > premium (list) - 2d matrix representing the premium squares on
            the Scrabble board; '*' : no bonus, 'd' double letter,
            't' : triple letter, 'D' double word, 'T' : triple word
        > words_of_length (dict) - keys are word lengths, values are all
            words added with that length
    """

    def __init__(self, collection):
        super().__init__()
        self.board = None
        self.minor = None
        self.rack = ""
        self.values = VALUE_STANDARD
        self.premium = PREMIUM_STANDARD
        self.words_of_length = defaultdict(set)
        self._setup(collection)

    def _setup(self, collection):
        """Upload the Dictionary words from a text file 'filename'. When
        solving, the words from this text file will be used for spell
        checking."""
        self.clear_words()
        for word in self._validate_collection(collection):
            if not word.isalpha():
                continue
            self.add_word(word)
            self.words_of_length[len(word)].add(word)

    def solve(self, board, rack):
        """
            Solves the given Scrabble board, using the tiles in 'rack'.

            Parameters:
            > board (list) - 2d matrix representing a Scrabble board, all
                letter tiles in upper case, blanks all lower case, empty
                spaces can be "", "*" or None
            > rack (list) - letters of the tiles on the rack, blanks are
                represented by a "#"

            Returns:
            > solutions (list) - list of solutions, each solution is of
                the form (word, x, y, orientation, score)

            IMPORTANT NOTE:
            The algorithm used will not use blanks in place of tiles it
            already has. Blanks will only be to make a word if it has to
        """
        matrix = self._validate_board(board)
        rack = self._validate_rack(rack)
        solutions = set()

        # Gather horizontal word solutions
        for word, x, y, score in self._horizontal_solve(matrix, rack):
            solutions.add((word, x, y, False, score))

        # Gather vertical word solutions
        for word, y, x, score in self._horizontal_solve(matrix.transpose(), rack):
            solutions.add((word, x, y, True, score))

        return sorted(solutions, key=lambda x: x[-1], reverse=True)

    def get_score(self, board, rack, attempt):
        """
            Returns the points earned from placing attempt on the given
            board with the given rack.

            Parameters:
            > board (list) - 2d matrix representing a Scrabble board, all
                letter tiles in upper case, blanks all lower case, empty
                spaces can be "", "*" or None
            > rack (list) - letters of the tiles on the rack, blanks are
                represented by a "#"
            > attempt (tuple) - information of the attempt of the form:
                (word, x, y, orientation)
        """
        board = self._validate_board(board)
        rack = self._validate_rack(rack)
        (word, x, y, orientation) = self._validate_attempt(attempt)

        if not orientation:
            self._prepare(np.array(board), rack)
        else:
            self._prepare(np.array(board).transpose(), rack)
            x, y = y, x

        placements, adjacent = [], False
        for n in range(len(word)):
            tile = self._get_tile(x+n, y)
            if tile.isalpha():
                adjacent = True
            elif not tile:
                break
            elif tile == "*":
                if len(placements) == len(self.rack):
                    break
                placements.append(n)
                if not adjacent:
                    adjacent = self._is_vertically_adjacent(x+n, y)
        else:
            if not adjacent:
                return -1
            for (_, score) in self._yield_solutions([word], x, y, placements):
                return score
        return -1

    def _horizontal_solve(self, board, rack):
        """
            Generator function which yields horizontal word solutions,
            for the given Scrabble board, using the tiles in 'rack'.

            Parameters:
            > board (list) - 2d matrix representing a Scrabble board, all
                letter tiles in upper case, blanks all lower case
            > rack (str) - letters of the tiles on the rack, blanks are
                represented by a "#"

            Yields:
            > (word, x, y, score)
                word (str) : the whole main word created
                x (int) : the x coordinate of the beginning of the word
                y (int) : the y coordinate of the beginning of the word
                score (int) : the points gained by creating the word
        """
        self._prepare(board, rack)
        for x, y in np.ndindex(self.board.shape):
            for word, score in self._horizontal_solutions(x, y):
                yield (word, x, y, score)

    def _horizontal_solutions(self, x, y):
        """
            Yields the possible words and scores given from placing a
            word in beginningin the position x, y.

            Parameters
            > x (int) - x position of the words
            > y (int) - y position of the words

            Yields:
            > (word, score) - word: str, score: int
        """

        # Not interested in positions with a preceding tile to the left
        if self._get_tile(x-1, y).isalpha():
            return

        # placements : list of indices of tiles placed since the first
        # requirements : letter, position pairs, words must have to fit
        # adjacent : is the current tile stream adjacent to placed tiles
        placements, requirements, adjacent = [], [], False

        # Streams tiles horizontally from (x, y), yielding all solutions
        for n in range(self.board.shape[0]):
            tile = self._get_tile(x+n, y)
            if tile.isalpha():
                requirements.append((tile, n))
            elif requirements or adjacent:
                words = self._find_words(n, requirements)
                for word, score in self._yield_solutions(
                        words, x, y, placements
                    ):
                    yield (word, score)
            if not tile:
                break
            elif tile == "*":
                if len(placements) == len(self.rack):
                    break
                placements.append(n)
                if not adjacent:
                    adjacent = self._is_vertically_adjacent(x+n, y)

    def _yield_solutions(self, words, x, y, placements):
        """
            Yields the words and scores given from placing the word in
            words in the given x, y position, given that new tiles are
            being placed with indices in placements.

            Parameters
            > words (iterable) - possible words to be checked
            > x (int) - x position of the words
            > y (int) - y position of the words
            > placements (lst) - indices of tiles placed to make the
                words

            Yields
            > (word, score) - word: str, score: int
        """
        for word in words:
            rack_tiles = self._get_rack_tiles(self.rack, word, placements)
            if not rack_tiles:
                continue
            score = 0
            placed = [(rack_tiles[n], pos) for n, pos in enumerate(placements)]
            for let, pos in placed:
                if not let in self.minor[y][x + pos]:
                    break
                if let.islower():
                    word = word[:pos] + let + word[pos+1:]
                score += self.minor[y][x + pos][let]
            else:
                bingo = (len(placements) == 7)
                score += self._evaluate_major_score(word, x, y, bingo)
                yield (word, score)

    def _get_minor_scores(self):
        """Returns the points gained from any vertical words. The return
        value is a 2d matrix representing positions on self.board. The
        entries are dicts, keys are possible tiles which can be placed
        and that position, values are the points earned."""
        minor_scores = np.full(self.board.shape, dict())
        for x, y in np.ndindex(self.board.shape):
            minor_scores[x][y] = {}
        minor_shape = (self.board.shape[0], self.board.shape[1] + 1)
        allow_all = dict.fromkeys(ascii_lowercase + ascii_uppercase, 0)

        # Iterate though positions vertically
        letters, pos = [], ()
        for x, y in np.ndindex(minor_shape):
            tile = self._get_tile(x, y)
            if not tile.isalpha() and pos:
                ind = letters.index("*")
                if letters == ["*"]:
                    minor_scores[pos[1]][pos[0]] = allow_all
                else:
                    for part in self._generate_minor_score(
                            letters, pos[0], pos[1], ind
                    ):
                        minor_scores[pos[1]][pos[0]].update(part)
                if tile == "*":
                    letters = letters[ind+1:]
            if not tile:
                letters, pos = [], ()
            else:
                letters.append(tile.upper())
            if tile == "*":
                pos = (x, y)

        return minor_scores

    def _generate_minor_score(self, letters, x, y, ind):
        """Yields key-value pairs of letters and the scores given by
        placing the tile at position x, y, having index ind in the word.
        """
        letter_mult = {"d" : 2, "t" : 3}.get(self.premium[y][x], 1)
        base_value = sum([self.values.get(letter, 0) for letter in letters])
        base_mult = {"D" : 2, "T" : 3}.get(self.premium[y][x], 1)

        # Get all words
        requirements = [
            (let, pos) for pos, let in enumerate(letters) if let != "*"
        ]
        words = self._find_words(len(letters), requirements)

        # Yield a solution for each word
        for letter in {word[ind] for word in words}:
            letter_value = self.values.get(letter, 0)
            yield {
                letter : base_mult * (base_value + letter_mult * letter_value),
                letter.lower() : base_mult * base_value
            }

    def _prepare(self, board, rack):
        """Prepares for calculating scores."""
        self.board = board
        self.minor = self._get_minor_scores()
        self.rack = "".join(rack)

    def _find_words(self, length, requirements):
        """Returns the set of all words of length 'length', meeting the
        criteria set in the list 'requirements'. For every tuple
        ("?", n) in 'requirements', the word must contain the letter "?"
        at index 'n'. If 'requirements' is empty, all words of length
        'length' are returned."""
        if not requirements:
            return self.words_of_length[length]
        set_list = [self.lookup(length, let, pos) for let, pos in requirements]
        return set.intersection(*set_list)

    def _is_vertically_adjacent(self, x, y):
        """Returns whether the given (x, y) position on self.board in
        on the centre of the board or has a tile above or below it."""
        if x == (self.board.shape[1]-1)/2 and y == (self.board.shape[0]-1)/2:
            return True
        return self._get_tile(x, y-1).isalpha() or self._get_tile(x, y+1).isalpha()

    def _evaluate_major_score(self, word, x, y, bingo):
        """"Returns the points gained for spelling the given word
        horizontally at position (x, y). Points gained from minor words
        are not added."""
        value_total, word_mult = 0, 1
        bingo_bonus = 50 if bingo else 0
        for n, letter in enumerate(word):
            letter_mult = 1
            if self._get_tile(x+n, y) == "*":
                letter_mult = {"d" : 2, "t" : 3}.get(self.premium[y][x+n], 1)
                word_mult *= {"D" : 2, "T" : 3}.get(self.premium[y][x+n], 1)
            value_total += self.values.get(letter, 0) * letter_mult
        return value_total * word_mult + bingo_bonus

    def _get_tile(self, x, y):
        """Returns the letter of the tile on the board at position
        (x, y). If the position is free, '*' is returned. If (x, y)
        does not exist on the board, "" is returned."""
        if not (0 <= x < self.board.shape[1] and 0 <= y < self.board.shape[0]):
            return ""
        tile = self.board[y][x]
        if tile in ["", "*", None]:
            return "*"
        return tile

    @staticmethod
    def _get_rack_tiles(rack, word, placements):
        """Returns the word tiles used to place the word."""
        placed_tiles = [word[pos] for pos in placements]
        rack_tiles = ""
        for letter in placed_tiles:
            if letter in rack:
                rack = rack.replace(letter, "", 1)
                rack_tiles += letter
            elif "#" in rack:
                rack = rack.replace("#", "", 1)
                rack_tiles += letter.lower()
            else:
                return ""
        return rack_tiles

    @staticmethod
    def _validate_collection(collection):
        """Validate the collection parameter. Returns the words extracted."""

        # Return collection if it is already a list/set
        if isinstance(collection, (list, set)):
            return [word.upper() for word in collection]

        # Opens file and return lists of words in the file
        if isinstance(collection, str):
            with open(collection, "r") as file_text:
                text = file_text.read()
            return re.findall(r"[\w']+", text.upper())

        # Else an incorrect type has been received
        raise TypeError(
            "invalid value for 'collection' parameter. "
            "Expected list, set or str, received '%s'." % type(collection).__name__
        )

    @staticmethod
    def _validate_board(board):
        """Validate the board parameter. Returns the board matrix."""

        # Check board is a list
        if not isinstance(board, list):
            raise TypeError(
                "invalid value for 'board' parameter. "
                "Expected list, received '%s'." % type(board).__name__
            )

        # Check board is a list of length 15
        rows = len(board)
        if rows != 15:
            raise ValueError(
                "invalid value for 'board' parameter. "
                "Expected 15 rows, received %i.." % rows
            )

        # Check board is a list of lists
        element_lengths = set()
        for index, element in enumerate(board):
            if not isinstance(element, list):
                raise TypeError(
                    "invalid value for 'board' parameter at index %i. "
                    "Expected list, received '%s'." % (index, type(element).__name__)
                )
            element_lengths.add(len(element))

        # Check all lists are the same size
        if element_lengths != {15}:
            raise ValueError(
                "invalid value for 'board' parameter. "
                "Not all rows are 15 tiles long."
            )

        # Check each element in the lists are strings
        for index, element in enumerate(board):
            for tile in element:
                if not isinstance(tile, str):
                    raise TypeError(
                        "invalid value for 'board' parameter at index %i. "
                        "Not all elements are str, found '%s'" % (index, type(tile).__name__)
                    )
            board[index] = [
                (tile if tile.isalpha() else "*")
                for tile in element
            ]

        # Check that each element of board is a single character
        for index, element in enumerate(board):
            for letter in element:
                if len(letter) != 1:
                    raise ValueError(
                        "invalid value for 'board' parameter at index %i. "
                        "Cannot convert '%s' into a letter." % (index, letter)
                    )

        return np.array(board)

    @staticmethod
    def _validate_rack(rack):
        """Validate the rack parameter. Returns the rack."""

        # Check that rack is either a list, set or string
        if isinstance(rack, list):
            return rack
        if isinstance(rack, set):
            return list(rack)
        if isinstance(rack, str):
            return rack.split()
        raise TypeError(
            "invalid value for 'rack' parameter. "
            "Expected list, set or str, received '%s'." % type(rack).__name__
        )

    @staticmethod
    def _validate_attempt(attempt):
        """Validate the attempt parameter. Returns the attempt."""

        # Check the tuple count
        try:
            (word, x, y, orientation) = attempt
        except ValueError as exc:
            raise exc

        # Check word is an alphabetically string
        if not isinstance(word, str):
            raise ValueError(
                "invalid value for 'attempt' parameter. "
                "The word value is not an alphabetical string."
            )
        if not word.isalpha():
            raise ValueError(
                "invalid value for 'attempt' parameter. "
                "The word value is not an alphabetical string."
            )

        # Check the x value
        if not isinstance(x, int):
            raise ValueError(
                "invalid value for 'attempt' parameter. "
                "The x value is not an int."
            )

        # Check the y value
        if not isinstance(y, int):
            raise ValueError(
                "invalid value for 'attempt' parameter. "
                "The y value is not an int."
            )

        # Check the y value
        if not isinstance(orientation, bool):
            raise ValueError(
                "invalid value for 'attempt' parameter. "
                "The orientation value is not a bool."
            )

        return attempt



from setuptools import setup


with open('README.md', encoding='utf-8') as readme:
    long_description = readme.read()

INSTALL_REQUIRE = ["numpy>=1.19.2", "pythonds>=1.2.1"]
TESTS_REQUIRE = ["twine>=3.2.0"]

setup(
    author="Christopher Malcolm",
    author_email="chris.c.malcolm.96@gmail.com",
    name="wordsolver",
    version="1.0.1",
    description="A package to solve word games",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/chrismalcolm/solver",
    packages=["wordsolver"],
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=INSTALL_REQUIRE,
    test_require=TESTS_REQUIRE
)
